#include <TApplication.h>
#include <Rioplus.h>
#include <TLorentzVector.h>
#include <TTree.h>
#include <TLegend.h>
#include <TFile.h>
#include <TKey.h>
#include <TH1.h>
#include <TH2.h>
#include <TH3.h>
#include <TGraph.h>
#include <TCanvas.h>
#include <TStyle.h>
#include <TSystem.h>
#include <TEntryList.h>
#include <TROOT.h>
#include <TFormula.h>
#include <TClass.h>

#include <TFrameFormat.h>
#include <TCli.h>

using namespace std;

template <typename T>
std::vector<T> intersection(std::vector<T> v1, std::vector<T> v2)
{
        std::vector<T> v3;

        std::sort(v1.begin(), v1.end());
        std::sort(v2.begin(), v2.end());

        std::set_intersection(v1.begin(),v1.end(),
                                v2.begin(),v2.end(),
                                back_inserter(v3));
        return v3;
}

int main(int argc, char **argv)
{
        std::vector<TString> ff;
        Bool_t bTOC;
        bool bDump, bDraw;

        TString options;
        int compression;
        int dataQuality;

        TString select;
        std::vector<int> globalIDs;
        std::vector<double> gps;

        double gpsMin, gpsMax;
        double minDuration, maxDuration;
        
        std::vector<int> channels;
        std::vector<int> groups;
        
        int skip, max;

        TString outputFile;
        bool bMore, bHisto;
        bool duplicates;
        std::vector<TString> dataClassStr;

        TString tags;
        int decimate, fs;
        bool bOutputTree;
        int outputFrameLength;
        bool bOverride;

        TCli& cmdline = TCli::Instance(&argc, argv, "verbose multithread");

                cmdline.AddMessage("This script is used to compute trending curves from frame format file..");
                cmdline.AddOption("--options", "Custom chain options", options);
                cmdline.AddOption("--compression", "Compression level (0..255)", compression, 0);
                cmdline.AddOption("--duplicates", "Consider duplicates", duplicates);
                cmdline.AddOption("--data-quality", "Data quality requirements", dataQuality, 0);

                cmdline.AddOption("--channel-tags|--tags   |--name", "Only display specific structure names (you can use wildcard, multiple tags and anti-tags)", tags, "*");
                cmdline.AddOption("--adc-number  |--adc|-c", "Only display specific channels (use blank space as separator ` `; NB: Only for ADC)", channels);
                cmdline.AddOption("--adc-group   |--group  |-g", "Only display specific groups (use blank space as separator ` `; NB: Only for ADC)", groups);
                cmdline.AddOption("--adc-decimate|--decimate", "Decimate ADC channel values by a given factor (e.g. 1,2,3,4)", decimate, 1);

                cmdline.AddOption("--frame", "Select specific global frame IDs (range or list)", globalIDs);
                cmdline.AddOption("--gps", "Select specific GPS times (range or list)", gps);
                cmdline.AddOption("--gps-min|--gpsmin", "Minimal permitted GPS value", gpsMin, NAN);
                cmdline.AddOption("--gps-max|--gpsmax", "Maximum permitted GPS value", gpsMax, NAN);
                cmdline.AddOption("--min-duration", "Select frame with minimal duration", minDuration);
                cmdline.AddOption("--max-duration", "Select frame with maximal duration", maxDuration);
                cmdline.AddOption("--select", "Select specific frame (either range, list, or formula [x = global ID], [y = GPStime], [z = duration], and [t = data quality])", select, "");

                cmdline.AddOption("--output-gps", "Output tree structure including truncated GPS time information", bOutputTree);
                cmdline.AddOption("--override", "Override existing files", bOverride);

                cmdline.AddArgument(outputFile, "output.gwf");
                cmdline.AddArgument(fs, "sample_rate");
                cmdline.AddVariadic(ff, "test.gwf", 1);

        
        if(!cmdline.Usage() ) return 1;

        if(bOverride) options += "recreate";
        if(bOutputTree) options += " tree";
        
        std::vector<FrameFormat::FrType> dataClass;
                                        //  dataClass.push_back(FrameFormat::FrType::FrAdcData); // optimization removed for dev purpose. (@TODO: uncomment after that)

        TString fName = ROOT::IOPlus::Helpers::Implode(" ", ff);
        FrameFormat::TFrameChain f(fName.Data(), options + TString(duplicates ? "d" : ""), tags, dataClass, decimate, groups, channels);

        if (f.GetNFiles() == 1) bMore = true;

             if(gPrint->IsDebugEnabled((int) TCli::Verbosity::VERBOSITY_DEBUG        )) f.SetVerbosity(FrameFormat::Verbosity::ALL);
        else if(gPrint->IsDebugEnabled((int) TCli::Verbosity::VERBOSITY_VERY_VERBOSE )) f.SetVerbosity(FrameFormat::Verbosity::DEBUG);
        else if(gPrint->IsDebugEnabled((int) TCli::Verbosity::VERBOSITY_VERBOSE      )) f.SetVerbosity(FrameFormat::Verbosity::INFO);
        else f.SetVerbosity(FrameFormat::Verbosity::NORMAL);

        if( TPrint::ErrorIf(!f.IsOpen(), __METHOD_NAME__, "Failed to open `%s`", fName.Data()) )
                return 1;

        //
        // Select and save data
        std::vector<int> selection = f.Select(select.EqualTo("") ? "true" : select);

        // Select data quality
        if(dataQuality != 0) {
        
                std::vector<int> selectionDQ = f.SelectByDQ(dataQuality);
                selection = intersection(selection, selectionDQ);
        }
        
        if(gps.size() > 0) { 

                std::vector<int> selectionGPS;
                switch(gps.size()) {

                        case 0: break;

                        case 1: selectionGPS = f.SelectByGPS(gps);
                        break;

                        case 2: selectionGPS = f.SelectByGPS(gps[0], gps[1]);
                        break;

                        default: selectionGPS = f.SelectByGPS(gps);
                }

                selection = intersection(selection, selectionGPS);
        } 

        if(!ROOT::IOPlus::Helpers::IsNaN(gpsMin) || !ROOT::IOPlus::Helpers::IsNaN(gpsMax)) {

                std::vector<int> selectionGPS = f.SelectByGPS(gpsMin, gpsMax);
                selection = intersection(selection, selectionGPS);
        }
        
        if(maxDuration > 0) { 

                std::vector<int> selectionDuration = f.SelectByMaxDuration(maxDuration);
                                 selection = intersection(selection, selectionDuration);
        } 

        if(minDuration > 0) { 

                std::vector<int> selectionDuration = f.SelectByMinDuration(minDuration);
                                 selection = intersection(selection, selectionDuration);
        } 

        if (globalIDs.size()) {

                std::vector<int> selectionFrame;
                switch(globalIDs.size()) {

                        case 0: break;

                        case 1: selectionFrame = f.SelectByFrame(globalIDs);
                        break;

                        case 2: selectionFrame = f.SelectByFrame(globalIDs[0], globalIDs[1]);
                        break;

                        default: selectionFrame = f.SelectByFrame(globalIDs);
                }

                selection = intersection(selection, selectionFrame);
        }

        return f.Trend(outputFile, fs, options, selection) == NULL;
}
